#include <windows.h>
#include <stdio.h>
#include "args.h"

void print_help() {
    printf("Usage:\n"
           "  screcover [-<switches>] <source file> <target file>\n"
           "Switches:\n"
           "  -b <blocksize>      sets physical block size for media. 2048 is default.\n"
           "  -o[b|k|m] <offset>  offset in bytes (or optionally in blocks, KBytes or MBytes)\n"
           "                      to start recovery from\n"
           "  -c[b|k|m] <count>   number of bytes (or optionally in blocks, KBytes or MBytes)\n"
           "                      to recover\n"
           "  -r <retry>          number of retry attempts. 3 is default\n"
           "  -l <logfile>        log unreliable and bad blocks to <logfile>\n"
           "  -?                  display this help\n"
           "Note:\n"
           "  1 KByte = 1024 bytes\n"
           "  1 MByte = 1024 KBytes = 1024*1024 bytes\n"
           "  byte offset or byte count is specified, it will be rounded to blocksize\n"
           "Examples:\n"
           "  screcover -o 5000000 X:\\Video.dat E:\\Temp\\Restored_Film.dat\n"
           "    will copy data starting from byte offset 5000000 to end of file"
           "  screcover -om 1300 -cm 1000 X:\\Video.dat E:\\Temp\\Restored_Film.dat\n"
           "    will copy not more than 1000Mb of data starting from byte offset 1300Mb"
          );
    exit(0);
}

__int64
get_file_size(
    HANDLE h
    )
{
    LONG hsz = 0;
    LONG lsz;

    lsz = SetFilePointer(h, 0, &hsz, FILE_END);
    return (((__int64)hsz) << 32) | lsz;
}

__int64
set_file_pointer(
    HANDLE h,
    __int64 sz
    )
{
    ULONG hsz = (ULONG)(sz >> 32);
    ULONG lsz = (ULONG)sz;

    lsz = SetFilePointer(h, lsz, (PLONG)&hsz, FILE_BEGIN);
    return (((__int64)hsz) << 32) | lsz;
}

int
set_file_size(
    HANDLE h,
    __int64 sz
    )
{
    set_file_pointer(h, sz);
    return SetEndOfFile(h);
}

void main (void) {
    int i, j;
    WCHAR a;
    ULONG bs = 2048;
    ULONG retry = 3;
    ULONG r;

    PWCHAR in_nameW = NULL;
    PWCHAR out_nameW = NULL;
    PWCHAR log_nameW = NULL;

    PCHAR in_nameA = NULL;
    PCHAR out_nameA = NULL;
    PCHAR log_nameA = NULL;
    
    int break_sw;
    int block_offs = 0;
    int block_counter = 0;
    __int64 offs = 0;
    __int64 counter = 0;
    __int64 f_size;
    __int64 f_ptr;
    __int64 bads = 0;
    __int64 susp = 0;
    __int64 copy_count = 0;
    char* buff;
    ULONG read_bytes;
    BOOLEAN ok;
    __int64 p;

    WCHAR* CmdLineW;
    PWCHAR* argvW;
    CHAR*  CmdLineA;
    PCHAR* argvA;
    int argc;

    CHAR LogBuffer[128];

    HANDLE in_h;
    HANDLE out_h;
    HANDLE log_h = NULL;

    printf("Scratched File Copy utility v1.4 (c) by Alter, 2004\n"
           "Home site http://www.alter.org.ua\n");

    if(GetVersion() & 0x80000000) {
//        printf("Sorry, for NT family only\n");
//        exit(1);
        CmdLineA = GetCommandLineA();
        argvA = CommandLineToArgvA(CmdLineA, &argc);
#define argv     argvA
#define in_name  in_nameA
#define out_name out_nameA
#define log_name log_nameA
#define xcslen   strlen
#define xscanf   sscanf
#define xprintf  printf
#define CreateFileX   CreateFileA
#define XT(a)    a

        #include "parse_args.cpp.h"

    } else {
        CmdLineW = GetCommandLineW();
        argvW = CommandLineToArgvW(CmdLineW, &argc);
#pragma warning(push)               
#pragma warning(disable:4005)
//#pragma warning(disable:6311)
#define argv     argvW
#define in_name  in_nameW
#define out_name out_nameW
#define log_name log_nameW
#define xcslen   wcslen
#define xscanf   swscanf
#define xprintf  wprintf
#define CreateFileX   CreateFileW
#define XT(a)    L##a
#pragma warning(pop)

        #include "parse_args.cpp.h"
    
    }

    buff = (char*)GlobalAlloc(GMEM_DISCARDABLE, bs);
    if(!buff) {
        printf("Can't alloc %d bytes for temporary buffer\n"
               "exiting...\n", bs);
        exit (-2);
    }

    f_size = get_file_size(in_h);
    if(!f_size) {
        if(GetVersion() & 0x80000000) {
            printf("Empty source file %s\n"
                   "exiting...\n", out_nameA);
        } else {
            wprintf(L"Empty source file %s\n"
                   L"exiting...\n", out_name);
        }
        exit (-4);
    }

    if(!set_file_size(out_h, f_size)) {
        if(GetVersion() & 0x80000000) {
            printf("Can't set size of target file %s to %I64d bytes\n"
                   "exiting...\n", out_nameA, f_size);
        } else {
            wprintf(L"Can't set size of target file %s to %I64d bytes\n"
                   L"exiting...\n", out_name, f_size);
        }
        exit (-5);
    }

    // align offs and counter
    offs = offs & ~((unsigned __int64)(bs-1));
    counter = (counter + bs - 1) & ~((unsigned __int64)(bs-1));

    f_ptr = offs;
    if(block_offs) {
        f_ptr *= bs;
    }
    if(block_counter) {
        counter *= bs;
    }

    if(GetVersion() & 0x80000000) {
        printf("%s -> %s\n%I64d bytes\n",
               in_nameA, out_nameA, f_size);
    } else {
        wprintf(L"%s -> %s\n%I64d bytes\n",
               in_name, out_name, f_size);
    }
    printf("Starting from offset %I64d bytes\n",
           offs);
    if(counter) {
        printf("Recovery range %I64d bytes\n",
               counter);
    }

    p = f_size/50;
    if(!p) p = 1;

    while(f_ptr <= f_size-bs) {
        if(counter && copy_count >= counter) {
            printf("Copied %I64d bytes, %I64d unreliable, %I64d in bad blocks\n", copy_count, susp, bads);
            exit(0);
        }
        if((f_ptr % p) < bs)
            printf(".");
        memset(buff, 0, bs);
        ok = FALSE;
        for(r=0; r<retry; r++) {
            set_file_pointer(in_h, f_ptr);
            set_file_pointer(out_h, f_ptr);
            read_bytes = 0;
            if(!ReadFile(in_h, buff, bs, &read_bytes, NULL) ||
                read_bytes != bs) {
                read_bytes = GetLastError();
                if(!r) {
                    printf("Can't read block at offset %I64d\n", f_ptr);
                }
            } else {
                if(r) {
                    printf("Read on attempt #%d\n", r+1);
                }
                WriteFile(out_h, buff, bs, &read_bytes, NULL);
                if(read_bytes != bs) {
                    printf("Can't write block at offset %I64d\n", f_ptr);
                } else {
                    ok = TRUE;
                }
                break;
            }
        }
        if(!ok) {
            bads += bs;
            if(log_h) {
                sprintf(LogBuffer, "bad at %I64d size %d\n", f_ptr, bs);
                WriteFile(log_h, LogBuffer, strlen(LogBuffer), &read_bytes, NULL);
            }
        } else
        if(r) {
            susp += bs;
            if(log_h) {
                sprintf(LogBuffer, "unreliable at %I64d size %d\n", f_ptr, bs);
                WriteFile(log_h, LogBuffer, strlen(LogBuffer), &read_bytes, NULL);
            }
        }
        f_ptr += bs;
        copy_count += bs;
    }

    printf("\n");
    bs = (ULONG)(min(bs, f_size - f_ptr));
    if(!bs || (counter && copy_count >= counter)) {
        printf("Copied %I64d bytes, %I64d unreliable, %I64d in bad blocks\n", copy_count, susp, bads);
        exit(0);
    }
    memset(buff, 0, bs);
    ReadFile(in_h, buff, bs, &read_bytes, NULL);
    if(read_bytes != bs) {
        printf("Can't read block at offset %I64d\n", f_ptr);
        bads += bs;
    }
    WriteFile(out_h, buff, bs, &read_bytes, NULL);
    if(read_bytes != bs) {
        printf("Can't write block at offset %I64d\n", f_ptr);
    }

    printf("Copied %I64d bytes, %I64d unreliable, %I64d in bad blocks\n", copy_count, susp, bads);
    exit(0);
}
